/**************************************************************************
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * gPXE driver datapath for Solarflare network cards
 *
 * Written by Michael Brown <mbrown@fensystems.co.uk>
 * Written by Steve Hodgson <shodgson@solarflare.com>
 *
 * Copyright Fen Systems Ltd. 2005
 * Copyright Level 5 Networks Inc. 2005
 * Copyright 2006-2009 Solarflare Communications Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 (GPL), incorporated herein by
 * reference.  Drivers based on or derived from this code fall under
 * the GPL and must retain the authorship, copyright and license
 * notice.
 *
 **************************************************************************
 */
FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <byteswap.h>
#include <gpxe/io.h>
#include <gpxe/pci.h>
#include <gpxe/malloc.h>
#include <gpxe/iobuf.h>
#include <gpxe/netdevice.h>
#include "efx.h"
#include "efx_bitfield.h"
#include "efx_regs.h"

#include <config/general.h>
#ifdef EFX_ENABLE_NVLOG
#include "efx_nvram.h"
#include "mc_driver_pcol.h"
#endif

/***************************************************************************
 *
 * Constants and macros
 *
 ***************************************************************************/
extern int __invalid_queue_size;
#define FQS(_prefix, _x)				\
	( ( (_x) == 512 ) ? _prefix ## _SIZE_512 :	\
	  ( ( (_x) == 1024 ) ? _prefix ## _SIZE_1K :	\
	  ( ( (_x) == 2048 ) ? _prefix ## _SIZE_2K :	\
	  ( ( (_x) == 4096) ? _prefix ## _SIZE_4K :	\
	  __invalid_queue_size ) ) ) )

/*******************************************************************************
 *
 *
 * Low-level hardware access
 *
 *
 *******************************************************************************/ 
static void
efx_write_sram ( struct efx_nic *efx, efx_qword_t *value,
		 unsigned int index )
{
	unsigned int reg;

	reg = FR_BZ_BUF_FULL_TBL + ( index * sizeof ( *value ) );
	EFX_REGDUMP ( "Writing SRAM register %x with " EFX_QWORD_FMT "\n",
		      reg, EFX_QWORD_VAL ( *value ) );

	_efx_writel ( efx, value->u32[0], reg + 0  );
	_efx_writel ( efx, value->u32[1], reg + 4  );
	wmb();
}

/*******************************************************************************
 *
 *
 * SRAM
 *
 *
 *******************************************************************************/

static void
efx_free_special_buffer ( void *p )
{
	/* We don't bother cleaning up the buffer table entries -
	 * we're hardly limited */
	free_dma ( p, EFX_BUF_ALIGN );
}

static void *
efx_alloc_special_buffer ( struct efx_nic *efx, int bytes,
			   struct efx_special_buffer *entry )
{
	void* buffer;
	int remaining;
	efx_qword_t buf_desc;
	unsigned long dma_addr;

	/* Allocate the buffer, aligned on a buffer address boundary */
	buffer = malloc_dma ( bytes, EFX_BUF_ALIGN );
	if ( ! buffer )
		return NULL;

	/* Push buffer table entries to back the buffer */
	entry->id = efx->buffer_head;
	entry->dma_addr = dma_addr = virt_to_bus ( buffer );
	assert ( ( dma_addr & ( EFX_BUF_ALIGN - 1 ) ) == 0 );

	remaining = bytes;
	while ( remaining > 0 ) {
		EFX_POPULATE_QWORD_3 ( buf_desc,
				       FRF_AZ_BUF_ADR_REGION, 0,
				       FRF_AZ_BUF_ADR_FBUF, dma_addr >> 12,
				       FRF_AZ_BUF_OWNER_ID_FBUF, 0 );

		efx_write_sram ( efx, &buf_desc, efx->buffer_head );

		++efx->buffer_head;
		dma_addr += EFX_BUF_ALIGN;
		remaining -= EFX_BUF_ALIGN;
	}

	EFX_TRACE ( "Allocated 0x%x bytes at %p backed by buffer table "
		    "entries 0x%x..0x%x\n", bytes, buffer, entry->id,
		    efx->buffer_head - 1 );

	return buffer;
}

/*******************************************************************************
 *
 *
 * RX
 *
 *
 *******************************************************************************/

static void
efx_build_rx_desc ( efx_rx_desc_t *rxd, struct io_buffer *iob )
{
	EFX_POPULATE_QWORD_3 ( *rxd,
			       FSF_AZ_RX_KER_BUF_SIZE, EFX_RX_BUF_SIZE,
			       FSF_AZ_RX_KER_BUF_REGION, 0,
			       FSF_AZ_RX_KER_BUF_ADDR,
			       virt_to_bus ( iob->data ) );
}

static void
efx_notify_rx_desc ( struct efx_nic *efx )
{
	struct efx_rx_queue *rxq = &efx->rxq;
	int ptr = rxq->write_ptr % EFX_RXD_SIZE;
	efx_dword_t reg;

	EFX_POPULATE_DWORD_1 ( reg, FRF_AZ_RX_DESC_WPTR_DWORD, ptr );
	efx_writel_page ( efx, &reg, 0, FR_AZ_RX_DESC_UPD_DWORD_P0 );
}

static void
efx_rxq_fill ( struct efx_nic *efx )
{
	struct efx_rx_queue *rxq = &efx->rxq;
	int fill_level = rxq->write_ptr - rxq->read_ptr;
	int space = EFX_NUM_RX_DESC - fill_level - 1;
	int pushed = 0;

	while ( space ) {
		int buf_id = rxq->write_ptr % EFX_NUM_RX_DESC;
		int desc_id = rxq->write_ptr % EFX_RXD_SIZE;
		struct io_buffer *iob;
		efx_rx_desc_t *rxd;

		assert ( rxq->buf[buf_id] == NULL );
		iob = alloc_iob ( EFX_RX_BUF_SIZE );
		if ( !iob )
			break;

		EFX_TRACE ( "pushing rx_buf[%d] iob %p data %p\n",
			    buf_id, iob, iob->data );

		rxq->buf[buf_id] = iob;
		rxd = rxq->ring + desc_id;
		efx_build_rx_desc ( rxd, iob );
		++rxq->write_ptr;
		++pushed;
		--space;
	}

	/* Push the ptr to hardware */
	if ( pushed > 0 ) {
		efx_notify_rx_desc ( efx );

		EFX_TRACE ( "pushed %d rx buffers to fill level %d\n",
			    pushed, rxq->write_ptr - rxq->read_ptr );
	}
}

void
efx_receive ( struct efx_nic *efx, unsigned int id, int len, int drop )
{
	struct efx_rx_queue *rxq = &efx->rxq;
	unsigned int read_ptr = rxq->read_ptr % EFX_RXD_SIZE;
	unsigned int buf_ptr = rxq->read_ptr % EFX_NUM_RX_DESC;
	struct io_buffer *iob;

	assert ( id == read_ptr );

	/* Pop this rx buffer out of the software ring */
	iob = rxq->buf[buf_ptr];
	rxq->buf[buf_ptr] = NULL;

	DBGCIO(efx, "popping rx_buf[%d] iob %p data %p with %d bytes %s %x\n",
	       id, iob, iob->data, len, drop ? "bad" : "ok", drop);

#ifdef EFX_ENABLE_NVLOG
	/* Start logging received packets only after we have sent one out. */
	if (efx->txq.write_ptr)
		efx_nvlog_packet(efx,
				 iob->data + efx->rx_prefix_size,
				 len - efx->rx_prefix_size);
#endif

	/* Pass the packet up if required */
	if ( drop )
		netdev_rx_err(efx->netdev, iob, EBADMSG);
	else {
		iob_put ( iob, len );
		netdev_rx ( efx->netdev, iob );
	}

	++rxq->read_ptr;
}

/*******************************************************************************
 *
 *
 * TX
 *
 *
 *******************************************************************************/

static void
efx_build_tx_desc ( efx_tx_desc_t *txd, struct io_buffer *iob )
{
	EFX_POPULATE_QWORD_2 ( *txd,
			       FSF_AZ_TX_KER_BYTE_COUNT, iob_len ( iob ),
			       FSF_AZ_TX_KER_BUF_ADDR, virt_to_bus ( iob->data ) );
}

static void
efx_notify_tx_desc ( struct efx_nic *efx )
{
	struct efx_tx_queue *txq = &efx->txq;
	int ptr = txq->write_ptr % EFX_TXD_SIZE;
	efx_dword_t reg;

	EFX_POPULATE_DWORD_1 ( reg, FRF_AZ_TX_DESC_WPTR_DWORD, ptr );
	efx_writel_page ( efx, &reg, 0, FR_AZ_TX_DESC_UPD_DWORD_P0 );
}

int
efx_transmit ( struct net_device *netdev, struct io_buffer *iob )
{
	struct efx_nic *efx = netdev_priv ( netdev );
	struct efx_tx_queue *txq = &efx->txq;
	int fill_level, space;
	efx_tx_desc_t *txd;
	int buf_id;

	fill_level = txq->write_ptr - txq->read_ptr;
	space = EFX_TXD_SIZE - fill_level - 1;
	if ( space < 1 )
		return -ENOBUFS;

	/* Save the iobuffer for later completion */
	buf_id = txq->write_ptr % EFX_TXD_SIZE;
	assert ( txq->buf[buf_id] == NULL );
	txq->buf[buf_id] = iob;

	DBGCIO(efx, "tx_buf[%d] for iob %p data %p len %zd\n",
	       buf_id, iob, iob->data, iob_len(iob));

#ifdef EFX_ENABLE_NVLOG
	efx_nvlog_packet(efx, iob->data, iob_len(iob));
#endif

	/* Form the descriptor, and push it to hardware */
	txd = txq->ring + buf_id;
	efx_build_tx_desc ( txd, iob );
	++txq->write_ptr;
	efx_notify_tx_desc ( efx );

	return 0;
}

static int
efx_transmit_done ( struct efx_nic *efx, int id )
{
	struct efx_tx_queue *txq = &efx->txq;
	unsigned int read_ptr, stop;

	/* Complete all buffers from read_ptr up to and including id */
	read_ptr = txq->read_ptr % EFX_TXD_SIZE;
	stop = ( id + 1 ) % EFX_TXD_SIZE;

	while ( read_ptr != stop ) {
		struct io_buffer *iob = txq->buf[read_ptr];
		assert ( iob );

		/* Complete the tx buffer */
		if ( iob )
			netdev_tx_complete ( efx->netdev, iob );
		txq->buf[read_ptr] = NULL;
		
		++txq->read_ptr;
		read_ptr = txq->read_ptr % EFX_TXD_SIZE;
	}

	return 0;
}

/*******************************************************************************
 *
 *
 * Event queues and interrupts
 *
 *
 *******************************************************************************/ 

static void
efx_clear_interrupts ( struct efx_nic *efx )
{
	efx_dword_t reg;

	/* read the ISR */
	efx_readl( efx, &reg, FR_BZ_INT_ISR0 );
}

/**
 * See if an event is present
 *
 * @v event		Event structure
 * @ret True		An event is pending
 * @ret False		No event is pending
 *
 * We check both the high and low dword of the event for all ones.  We
 * wrote all ones when we cleared the event, and no valid event can
 * have all ones in either its high or low dwords.  This approach is
 * robust against reordering.
 *
 * Note that using a single 64-bit comparison is incorrect; even
 * though the CPU read will be atomic, the DMA write may not be.
 */
static inline int
efx_event_present ( efx_event_t* event )
{
	return ( ! ( EFX_DWORD_IS_ALL_ONES ( event->dword[0] ) |
		     EFX_DWORD_IS_ALL_ONES ( event->dword[1] ) ) );
}

static void
efx_evq_read_ack ( struct efx_nic *efx )
{
	struct efx_ev_queue *evq = &efx->evq;
	efx_dword_t reg;

	EFX_POPULATE_DWORD_1 ( reg, FRF_AZ_EVQ_RPTR, evq->read_ptr );
	efx_writel_table ( efx, &reg, 0, FR_BZ_EVQ_RPTR );
}


static unsigned int
efx_handle_event ( struct efx_nic *efx, efx_event_t *evt )
{
	int ev_code, desc_ptr, len, packet_error;

	/* Decode event */
	ev_code = EFX_QWORD_FIELD ( *evt, FSF_AZ_EV_CODE );
	switch ( ev_code ) {
	case FSE_AZ_EV_CODE_TX_EV:
		desc_ptr = EFX_QWORD_FIELD ( *evt, FSF_AZ_TX_EV_DESC_PTR );
		efx_transmit_done ( efx, desc_ptr );
		break;
	
	case FSE_AZ_EV_CODE_RX_EV:
		desc_ptr = EFX_QWORD_FIELD ( *evt, FSF_AZ_RX_EV_DESC_PTR );
		len = EFX_QWORD_FIELD ( *evt, FSF_AZ_RX_EV_BYTE_CNT );
		/* The RX_EV_PKT_OK is set to 0 if RX_EV_IP_FRAG_ERR is set.
		 * But RX_EV_IP_FRAG_ERR indicates a fragmented IP protocol
		 * packet, not an error. Thus we can ignore RX_EV_IP_FRAG_ERR.
		 * Compute packet_ok the same way RX_EV_PKT_OK is but ignoring
		 * RX_EV_IP_FRAG_ERR.
		 *
		 * Note that bit 37 should not be used to compute packet_ok (see
		 * siena hdl/core/port/rxdp/rxp/riff_rctl.v looking for
		 * st1_ev_dat[37]) */
		packet_error = (evt->u64[0] &
			/* Bit 58 */
			((1ULL << FSF_CZ_RX_EV_PKT_NOT_PARSED_LBN) |
			 /* Bit 55-54 */
			 (1ULL << FSF_AZ_RX_EV_PAUSE_FRM_ERR_LBN) |
			 (1ULL << FSF_AZ_RX_EV_BUF_OWNER_ID_ERR_LBN) |
			 /* Bit 52-47 */
			 (1ULL << FSF_AZ_RX_EV_IP_HDR_CHKSUM_ERR_LBN) |
			 (1ULL << FSF_AZ_RX_EV_TCP_UDP_CHKSUM_ERR_LBN) |
			 (1ULL << FSF_AZ_RX_EV_ETH_CRC_ERR_LBN) |
			 (1ULL << FSF_AZ_RX_EV_FRM_TRUNC_LBN) |
			 (1ULL << FSF_AZ_RX_EV_TOBE_DISC_LBN))
			) >> 32;

		efx_receive ( efx, desc_ptr, len, packet_error );
		return 1;

	default:
		EFX_TRACE ( "Unknown event type %d\n", ev_code );
		break;
	}
	return 0;
}

void efx_poll ( struct net_device *netdev )
{
	struct efx_nic *efx = netdev_priv ( netdev );
	struct efx_ev_queue *evq = &efx->evq;
	efx_event_t *evt;
	int budget = 10;

	/* Read the event queue by directly looking for events
	 * (we don't even bother to read the eventq write ptr) */
	evt = evq->ring + evq->read_ptr;
	while (efx_event_present(evt) && (budget > 0)) {

		EFX_TRACE ( "Event at index 0x%x address %p is "
			    EFX_QWORD_FMT "\n", evq->read_ptr,
			    evt, EFX_QWORD_VAL ( *evt ) );

		budget -= efx_handle_event(efx, evt);

		/* Clear the event */
		EFX_SET_QWORD ( *evt );

		/* Move to the next event. We don't ack the event
		 * queue until the end */
		evq->read_ptr = ( ( evq->read_ptr + 1 ) % EFX_EVQ_SIZE );
		evt = evq->ring + evq->read_ptr;
	}

	/* Push more rx buffers if needed */
	efx_rxq_fill ( efx );

	/* Clear any pending interrupts */
	efx_clear_interrupts ( efx );

	/* Ack the event queue if interrupts are enabled */
	if ( efx->int_en )
		efx_evq_read_ack ( efx );
}

void efx_interrupts ( struct efx_nic *efx, int enabled, int force )
{
	efx_oword_t reg;

	EFX_POPULATE_OWORD_2 ( reg,
			       FRF_AZ_KER_INT_KER, force,
			       FRF_AZ_DRV_INT_EN_KER, 1 );
	efx_write ( efx, &reg, FR_AZ_INT_EN_KER );
	efx->int_en = enabled;

	/* Disarm the evq, then ack any pending interrupts */
	if ( ( efx->netdev->state & NETDEV_OPEN ) && !enabled ) {
		EFX_POPULATE_OWORD_2 ( reg,
				       FSF_AZ_EV_CODE,
				       FSE_AZ_EV_CODE_DRV_GEN_EV,
				       FSF_AZ_DRV_GEN_EV_MAGIC, 0 );
		efx_write ( efx, &reg, FR_AZ_DRV_EV );

		mdelay ( 1 );
		efx_clear_interrupts ( efx );
	}
}

void efx_irq ( struct net_device *netdev, int enable )
{
	struct efx_nic *efx = netdev_priv ( netdev );

	switch ( enable ) {
	case 0:
		efx_interrupts ( efx, 0, 0 );
		break;
	case 1:
		efx_interrupts ( efx, 1, 0 );
		efx_evq_read_ack ( efx );
		break;
	case 2:
		efx_interrupts ( efx, 1, 1 );
		break;
	}
}

/*******************************************************************************
 *
 *
 * Inititialization
 *
 *
 *******************************************************************************/
int efx_open ( struct net_device *netdev )
{
	struct efx_nic *efx = netdev_priv ( netdev );
	struct efx_ev_queue *evq = &efx->evq;
	struct efx_rx_queue *rxq = &efx->rxq;
	struct efx_tx_queue *txq = &efx->txq;
	efx_oword_t reg;
	efx_dword_t cmd;
	size_t bytes;
	int rc;

	/* Allocate the hardware event queue */
	bytes = sizeof ( efx_event_t ) * EFX_EVQ_SIZE;
	evq->ring = efx_alloc_special_buffer ( efx, bytes, &evq->entry );
	if ( evq->ring == NULL ) {
		rc = -ENOMEM;
		goto fail1;
	}
	memset ( evq->ring, 0xff, bytes );

	/* Allocate hardware receive queue */
	bytes = sizeof ( efx_rx_desc_t ) * EFX_RXD_SIZE;
	rxq->ring = efx_alloc_special_buffer ( efx, bytes, &rxq->entry );
	if ( rxq->ring == NULL ) {
		rc = -ENOMEM;
		goto fail2;
	}

	/* Allocate hardware transmit queue */
	bytes = sizeof ( efx_tx_desc_t ) * EFX_TXD_SIZE;
	txq->ring = efx_alloc_special_buffer ( efx, bytes, &txq->entry );
	if ( ! txq->ring ) {
		rc = -ENOMEM;
		goto fail3;
	}

	/* Inititialise pointers */
	evq->read_ptr = 0;
	txq->read_ptr = txq->write_ptr = 0;
	rxq->read_ptr = rxq->write_ptr = 0;

	/* Initialise timer table */
	EFX_POPULATE_OWORD_3 ( reg,
			       FRF_CZ_TIMER_Q_EN, 1,
			       FRF_CZ_HOST_NOTIFY_MODE, 0,
			       FRF_CZ_TIMER_MODE, FFE_CZ_TIMER_MODE_DIS );
	efx_write_table ( efx, &reg, 0, FR_BZ_TIMER_TBL );

	/* Push event queue */
	EFX_POPULATE_OWORD_3 ( reg,
			       FRF_AZ_EVQ_EN, 1,
			       FRF_AZ_EVQ_SIZE, FQS(FFE_AZ_EVQ, EFX_EVQ_SIZE),
			       FRF_AZ_EVQ_BUF_BASE_ID, evq->entry.id );
	efx_write_table ( efx, &reg, 0, FR_BZ_EVQ_PTR_TBL );

	/* Set interrupt moderation */
	EFX_ZERO_DWORD ( cmd );
	efx_writel_page ( efx, &cmd, 0, FR_BZ_TIMER_COMMAND_P0 );
	
	/* Push transmit queue */
	EFX_POPULATE_OWORD_9 ( reg,
			       FRF_AZ_TX_DESCQ_EN, 1,
			       FRF_AZ_TX_ISCSI_DDIG_EN, 0,
			       FRF_AZ_TX_ISCSI_DDIG_EN, 0,
			       FRF_AZ_TX_DESCQ_BUF_BASE_ID, txq->entry.id,
			       FRF_AZ_TX_DESCQ_EVQ_ID, 0,
			       FRF_AZ_TX_DESCQ_OWNER_ID, 0,
			       FRF_AZ_TX_DESCQ_SIZE, FQS(FFE_AZ_TX_DESCQ, EFX_TXD_SIZE),
			       FRF_AZ_TX_DESCQ_TYPE, 0 /* kernel queue */,
			       FRF_BZ_TX_NON_IP_DROP_DIS, 1 );
	efx_write_table ( efx, &reg, 0, FR_BZ_TX_DESC_PTR_TBL );

	EFX_POPULATE_OWORD_9 ( reg,
			       FRF_AZ_RX_ISCSI_DDIG_EN, 0,
			       FRF_AZ_RX_ISCSI_HDIG_EN, 0,
			       FRF_AZ_RX_DESCQ_BUF_BASE_ID, rxq->entry.id,
			       FRF_AZ_RX_DESCQ_EVQ_ID, 0,
			       FRF_AZ_RX_DESCQ_OWNER_ID, 0,
			       FRF_AZ_RX_DESCQ_SIZE, FQS(FFE_AZ_RX_DESCQ, EFX_RXD_SIZE),
			       FRF_AZ_RX_DESCQ_TYPE, 0 /* kernel queue */,
			       FRF_AZ_RX_DESCQ_JUMBO, 0,
			       FRF_AZ_RX_DESCQ_EN, 1 );
	efx_write_table ( efx, &reg, 0, FR_BZ_RX_DESC_PTR_TBL );

	/* Mask all fatal interrupts */
	EFX_ZERO_OWORD ( reg );
	efx_write ( efx, &reg, FR_AZ_FATAL_INTR_KER );

	/* Program INT_ADR_REG_KER */
	EFX_POPULATE_OWORD_1 ( reg, FRF_AZ_INT_ADR_KER,
			       virt_to_bus ( &efx->int_ker ) );
	efx_write ( efx, &reg, FR_AZ_INT_ADR_KER );

	/* Enable interrupts globally. Otherwise a subsequent driver
	 * could get a spurious interrupt when interrupts are enabled.
	 */
	EFX_POPULATE_OWORD_1 ( reg, FRF_AZ_DRV_INT_EN_KER, 1 );
	efx_write ( efx, &reg, FR_AZ_INT_EN_KER );

	/* Ack the eventq if interrupts are enabled */
	if ( efx->int_en )
		efx_evq_read_ack ( efx );

	/* Push receive buffers */
	efx_rxq_fill ( efx );

#ifdef EFX_ENABLE_NVLOG
	efx_nvlog_init(efx, MC_CMD_NVRAM_TYPE_MC_FW_BACKUP);
#endif

	return 0;

  fail3:
	efx_free_special_buffer ( rxq->ring );
  fail2:
	efx_free_special_buffer ( evq->ring );
  fail1:
	return rc;
}

void efx_close ( struct net_device *netdev )
{
	struct efx_nic *efx = netdev_priv ( netdev );
	struct efx_ev_queue *evq = &efx->evq;
	struct efx_rx_queue *rxq = &efx->rxq;
	struct efx_tx_queue *txq = &efx->txq;
	efx_oword_t cmd;
	int i;

	/* Flush the DMA queues */
	EFX_POPULATE_OWORD_2 ( cmd,
			       FRF_AZ_RX_FLUSH_DESCQ_CMD, 1,
			       FRF_AZ_RX_FLUSH_DESCQ, 0 );
	efx_write ( efx, &cmd, FR_AZ_RX_FLUSH_DESCQ );

	EFX_POPULATE_OWORD_2 ( cmd,
			       FRF_AZ_TX_FLUSH_DESCQ_CMD, 1,
			       FRF_AZ_TX_FLUSH_DESCQ, 0 );
	efx_write ( efx, &cmd, FR_AZ_TX_FLUSH_DESCQ );

	/* Wait for flush to complete */
	mdelay ( 100 );

	/* Remove descriptor rings */
	EFX_ZERO_OWORD ( cmd );
	efx_write_table ( efx, &cmd, 0, FR_BZ_RX_DESC_PTR_TBL );
	efx_write_table ( efx, &cmd, 0, FR_BZ_TX_DESC_PTR_TBL );
	efx_write_table ( efx, &cmd, 0, FR_BZ_EVQ_PTR_TBL );
	efx_write_table ( efx, &cmd, 0, FR_BZ_TIMER_TBL );

	/* Complete outstanding descriptors */
	for ( i = 0; i < EFX_NUM_RX_DESC; i++ ) {
		if ( rxq->buf[i] ) {
			free_iob ( rxq->buf[i] );
			rxq->buf[i] = NULL;
		}
	}

	for ( i = 0; i < EFX_TXD_SIZE; i++ ) {
		if ( txq->buf[i] ) {
			netdev_tx_complete ( efx->netdev, txq->buf[i] );
			txq->buf[i] = NULL;
		}
	}

	/* Remove software queues */
	efx_free_special_buffer ( txq->ring );
	txq->ring = NULL;

	efx_free_special_buffer ( rxq->ring );
	rxq->ring = NULL;

	efx_free_special_buffer ( evq->ring );
	evq->ring = NULL;

	/* Disable and ack interrupts */
	EFX_POPULATE_OWORD_1 ( cmd,
			       FRF_AZ_DRV_INT_EN_KER, 0 );
	efx_write ( efx, &cmd, FR_AZ_INT_EN_KER );

	mdelay ( 1 );
	efx_clear_interrupts ( efx );
}

/*******************************************************************************
 *
 *
 * VPD
 *
 *
 *******************************************************************************/ 

#define EXTRACT_BYTE( _byte, _field )					\
	( ( ( _byte ) >> _field ## _LBN ) & ( ( 1 << _field ## _WIDTH ) -1 ) )

#define MIN(_a, _b )				\
	( ( _a ) < ( _b ) ? ( _a ) : ( _b ) )

#define	TAG_TYPE_LBN 7
#define	TAG_TYPE_WIDTH 1
#define	TAG_TYPE_LARGE_ITEM 1
#define	TAG_TYPE_SMALL_ITEM 0

#define	TAG_SMALL_ITEM_NAME_LBN 3
#define	TAG_SMALL_ITEM_NAME_WIDTH 4
#define	TAG_SMALL_ITEM_SIZE_LBN 0
#define	TAG_SMALL_ITEM_SIZE_WIDTH 3

#define	TAG_LARGE_ITEM_NAME_LBN 0
#define	TAG_LARGE_ITEM_NAME_WIDTH 7

#define	TAG_NAME_END 0x0f
#define	TAG_NAME_ID_STRING 0x02
#define	TAG_NAME_VPD_R 0x10
#define	TAG_NAME_VPD_W 0x11

int efx_vpd ( const uint8_t *vpd, size_t size, enum sf_vpd_tag tag,
	      uint16_t keyword, uint8_t *buf, size_t buflen )
{
	unsigned int offset;
	uint16_t taglen;
	uint8_t chunk, type, name, keylen;

	offset = 0;
	while ( 1 ) {
		if ( offset >= size ) {
			DBG ( "overrun\n" );
			return -EFAULT;
		}

		chunk = vpd[offset];
		type = EXTRACT_BYTE ( chunk, TAG_TYPE );

		switch ( type ) {
		case TAG_TYPE_SMALL_ITEM:
			name = EXTRACT_BYTE ( chunk, TAG_SMALL_ITEM_NAME );
			taglen = EXTRACT_BYTE ( chunk, TAG_SMALL_ITEM_SIZE );
			++offset;
			break;

		case TAG_TYPE_LARGE_ITEM:
			name = EXTRACT_BYTE ( chunk, TAG_LARGE_ITEM_NAME );

			if ( 3 + offset >= size ) {
				DBG ( "VPD chunk %d is truncated\n", name );
				return -EFAULT;
			}
			taglen = ( vpd[offset + 1] |
				   ( ( uint16_t ) vpd[offset + 2] << 8 ) );
			offset += 3;
			break;

		default:
			return -EFAULT;
		}

		DBG ( "found tag %d offset 0x%x taglen 0x%x\n", name, offset, taglen );

		if ( offset + taglen >= size ) {
			EFX_LOG ( "VPD tag %d is oversize\n", type );
			return -EFAULT;
		}

		if ( name == TAG_NAME_END )
			break;
		else if ( name != tag ) {
			offset += taglen;
			continue;
		}

		if ( tag == VPD_ID ) {
			memcpy ( buf, vpd + offset, MIN ( buflen, taglen ) );
			return taglen;
		}

		while ( taglen > 0 ) {
			if ( taglen < 3 ||
			     taglen < 3 + ( keylen = vpd[offset + 2]) ) {
				DBG ( "VPD keyword %x%x is truncated\n",
				      vpd[offset], vpd[offset+1] );
				return -EFAULT;
			}

			DBG ( "Found VPD keyword %x%x length %d\n",
			      vpd[offset], vpd[offset+1], keylen );

			if ( memcmp ( &keyword, &vpd[offset], 2 ) == 0 ) {
				memcpy ( buf, vpd + offset + 3,
					 MIN ( buflen, keylen ) );
				return keylen;
			}

			offset += 3 + keylen;
			taglen -= 3 + keylen;
		}
	}

	DBG ( "Unable to find keyword\n" );

	return -ENOENT;
}
