/**************************************************************************
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * NVRAM driver for Solarflare network cards
 *
 * Copyright 2016 Solarflare Communications Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 (GPL), incorporated herein by
 * reference.  Drivers based on or derived from this code fall under
 * the GPL and must retain the authorship, copyright and license
 * notice.
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <gpxe/malloc.h>
#include <gpxe/if_ether.h>
#include <gpxe/if_arp.h>
#include <gpxe/iobuf.h>
#include <gpxe/ip.h>
#include <gpxe/ip6.h>
#include <gpxe/tcp.h>

#include "efx_common.h"
#include "efx_nvram.h"
#include "mc_driver_pcol.h"

#ifdef EFX_ENABLE_NVLOG
#include <gpxe/timer.h>
#endif

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define EFX_NVRAM_CHUNK 128

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/* For uninteresting packets we only log the headers. */
#define NVLOG_TRUNCATED_SIZE	(sizeof(struct ethhdr) + sizeof(struct iphdr) + \
				 sizeof(struct tcp_header))

struct efx_partition {
	uint32_t type;
	size_t   size;
	size_t   erasesize;
	size_t   readsize;
	size_t   writesize;
	int      update_ready;
};

struct efx_partition *efx_nvram_info(struct efx_nic *efx, uint32_t type)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_INFO_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_NVRAM_INFO_OUT_LEN);
	struct efx_partition *partition;
	size_t outlen;
	int rc;

	MCDI_SET_DWORD(inbuf, NVRAM_INFO_IN_TYPE, type);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_INFO,
				 inbuf, MC_CMD_NVRAM_INFO_IN_LEN,
				 outbuf, MC_CMD_NVRAM_INFO_OUT_LEN, &outlen,
				 false);

	if (rc || outlen < MC_CMD_NVRAM_INFO_OUT_LEN)
		return NULL;

	partition = malloc(sizeof *partition);
	if (!partition)
		return NULL;

	partition->type = type;
	partition->size = MCDI_DWORD(outbuf, NVRAM_INFO_OUT_SIZE);
	partition->erasesize = MCDI_DWORD(outbuf, NVRAM_INFO_OUT_ERASESIZE);
	partition->readsize = EFX_NVRAM_CHUNK;
	partition->writesize = EFX_NVRAM_CHUNK;
	partition->update_ready = 0;

	return partition;
}

int efx_partition_erasesize(struct efx_partition *partition)
{
	return partition ? (int)partition->erasesize : -1;
}

int efx_partition_size(struct efx_partition *partition)
{
	return partition ? (int)partition->size : -1;
}

void efx_partition_free(struct efx_partition *partition)
{
	if (partition)
		free(partition);
}

int efx_nvram_start(struct efx_nic *efx, struct efx_partition *partition)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_UPDATE_START_IN_LEN);
	int rc;

	MCDI_SET_DWORD(inbuf, NVRAM_UPDATE_START_IN_TYPE, partition->type);
	rc = efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_UPDATE_START,
				 inbuf, MC_CMD_NVRAM_UPDATE_START_IN_LEN,
				 NULL, 0, NULL, false);

	if (rc == 0)
		partition->update_ready = 1;

	return rc;
}

int efx_nvram_finish(struct efx_nic *efx, struct efx_partition *partition)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_UPDATE_FINISH_IN_LEN);

	partition->update_ready = 0;

	MCDI_SET_DWORD(inbuf, NVRAM_UPDATE_FINISH_IN_TYPE, partition->type);
	MCDI_SET_DWORD(inbuf, NVRAM_UPDATE_FINISH_IN_REBOOT, 0);
	return efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_UPDATE_FINISH,
				   inbuf, MC_CMD_NVRAM_UPDATE_FINISH_IN_LEN,
				   NULL, 0, NULL, false);
}

int efx_nvram_read(struct efx_nic *efx, struct efx_partition *partition,
		   size_t offset, size_t len, void *buf)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_READ_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_NVRAM_READ_OUT_LENMAX);
	uint8_t *data = buf;
	size_t outlen;
	int rc;

	while (len) {
		size_t chunk = MIN(len, partition->readsize);

		MCDI_SET_DWORD(inbuf, NVRAM_READ_IN_TYPE, partition->type);
		MCDI_SET_DWORD(inbuf, NVRAM_READ_IN_OFFSET, offset);
		MCDI_SET_DWORD(inbuf, NVRAM_READ_IN_LENGTH, chunk);

		rc = efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_READ,
					 inbuf, MC_CMD_NVRAM_READ_IN_LEN,
					 outbuf, MC_CMD_NVRAM_READ_OUT_LENMAX,
					 &outlen, false);

		if (rc)
			return rc;

		if (outlen < MC_CMD_NVRAM_READ_OUT_LEN(chunk))
			return -EMSGSIZE;

		memcpy(data, MCDI_PTR(outbuf, NVRAM_READ_OUT_READ_BUFFER),
		       chunk);

		len -= chunk;
		data += chunk;
		offset += chunk;
	}

	return 0;
}

int efx_nvram_erase_locked(struct efx_nic *efx, struct efx_partition *partition,
			   unsigned int start, size_t size)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_ERASE_IN_LEN);
	int rc;

	MCDI_SET_DWORD(inbuf, NVRAM_ERASE_IN_TYPE, partition->type);
	MCDI_SET_DWORD(inbuf, NVRAM_ERASE_IN_OFFSET, start);
	MCDI_SET_DWORD(inbuf, NVRAM_ERASE_IN_LENGTH, size);
	rc = efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_ERASE,
				 inbuf, MC_CMD_NVRAM_ERASE_IN_LEN,
				 NULL, 0, NULL, false);

	return rc;
}

int efx_nvram_erase(struct efx_nic *efx, struct efx_partition *partition,
		    unsigned int start, size_t size)
{
	int rc;

	rc = efx_nvram_start(efx, partition);
	if (rc)
		return rc;

	rc = efx_nvram_erase_locked(efx, partition, start, size);

	efx_nvram_finish(efx, partition);

	return rc;
}

int efx_nvram_write_locked(struct efx_nic *efx, struct efx_partition *partition,
			   size_t offset, size_t len, const void *buf)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_WRITE_IN_LEN(EFX_NVRAM_CHUNK));
	const uint8_t *data = buf;
	int rc;

	if (!partition->update_ready)
		return -EINVAL;

	while (len) {
		size_t chunk = MIN(len, partition->writesize);

		MCDI_SET_DWORD(inbuf, NVRAM_WRITE_IN_TYPE, partition->type);
		MCDI_SET_DWORD(inbuf, NVRAM_WRITE_IN_OFFSET, offset);
		MCDI_SET_DWORD(inbuf, NVRAM_WRITE_IN_LENGTH, chunk);

		memcpy(MCDI_PTR(inbuf, NVRAM_WRITE_IN_WRITE_BUFFER), data,
		       chunk);

		rc = efx->type->mcdi_rpc(efx, MC_CMD_NVRAM_WRITE, inbuf,
					 MC_CMD_NVRAM_WRITE_IN_LEN(chunk),
					 NULL, 0, NULL, false);

		if (rc)
			return rc;

		len -= chunk;
		data += chunk;
		offset += chunk;
	}

	return 0;
}

int efx_nvram_write(struct efx_nic *efx, struct efx_partition *partition,
		    size_t offset, size_t len, const void *buf)
{
	int rc;

	rc = efx_nvram_start(efx, partition);
	if (rc)
		return rc;

	rc = efx_nvram_write_locked(efx, partition, offset, len, buf);

	efx_nvram_finish(efx, partition);

	return rc;
}

#ifdef EFX_ENABLE_NVLOG

static struct {
	struct efx_partition *partition;
	u32 offset;
	u32 limit;
} efx_nvlog_state;

static const u32 efx_nvlog_magic = 0x70616370; /* Little-endian "pcap" */
static const u32 efx_nvlog_magic_reuse = 0x50414350; /* Little-endian "PCAP" */

int efx_nvlog_init(struct efx_nic *efx, int partition_type)
{
	struct efx_partition *partition;
	int blockcount;
	int blocksize;
	int block;
	int start;
	int rc;

	partition = efx_nvram_info(efx, partition_type);
	if (!partition)
		return -EINVAL;

	printf("Packet trace initialising...");

	blocksize = efx_partition_erasesize(partition);
	blockcount = efx_partition_size(partition) / blocksize;

	if (!blockcount) {
		rc = -ENOENT;
		goto fail;
	}

	if (blockcount == 1) {
		block = 0;
	} else {
		/* Find first block that doesn't have the magic. */
		int i;
		u32 magic;

		block = 0; /* Use block zero if we don't find one. */
		for (i = 0; i < blockcount; i++) {
			magic = 0;
			rc = efx_nvram_read(efx, partition, i * blocksize,
					    sizeof magic, &magic);

			if (rc == 0 && magic != efx_nvlog_magic) {
				block = i;
				break;
			}
		}

		/* If the next block has the magic, mark it as ready for
		 * re-use next time around.
		 * The re-use magic only clears bits from the normal magic,
		 * so it can be written without erasing the block.
		 */
		i = (block + 1) % blockcount;
		rc = efx_nvram_read(efx, partition, i * blocksize,
				    sizeof magic, &magic);

		if (rc == 0 && magic == efx_nvlog_magic) {
			/* Result of this is non-fatal. */
			efx_nvram_write(efx, partition, i * blocksize,
					sizeof efx_nvlog_magic,
					&efx_nvlog_magic_reuse);
		}
	}

	start = block * blocksize;

	printf(" using block %d of %d, size %d\n", block,
	       blockcount, blocksize);

	rc = efx_nvram_erase(efx, partition, start, blocksize);
	if (rc)
		goto fail;

	rc = efx_nvram_write(efx, partition, start, sizeof efx_nvlog_magic,
			     &efx_nvlog_magic);
	if (rc)
		goto fail;

	efx_nvlog_state.partition = partition;
	efx_nvlog_state.offset = start + sizeof efx_nvlog_magic;
	efx_nvlog_state.limit = start + blocksize;

	return 0;

fail:
	efx_partition_free(partition);
	printf("Packet trace init failed: %s\n", strerror(rc));
	return rc;
}

static int log_entire_packet(struct net_device *netdev, const void *data)
{
	int rc = 1;
#ifdef EFX_NVLOG_TCP_HDR_ONLY
	struct ll_protocol *ll_protocol = netdev->ll_protocol;

	/* Check link-layer protocol */
	if (ll_protocol->ll_proto == htons(ARPHRD_ETHER)) {
		const struct ethhdr *dptr = data;

		if (dptr->h_protocol == htons(ETH_P_IP)) {
			struct iphdr *iphdr = (struct iphdr *) ++dptr;

			if (iphdr->protocol == IP_TCP)
				rc = 0;
		} else if (dptr->h_protocol == htons(ETH_P_IPV6)) {
			struct ip6_header *ip6hdr =
				(struct ip6_header *) ++dptr;

			if (ip6hdr->nxt_hdr == IP_TCP)
				rc = 0;
		}
	}
#else
	netdev = netdev;	/* unused */
	data = data;		/* unused */
#endif
	return rc;
}

void efx_nvlog_packet(struct efx_nic *efx, const u8 *data, u32 data_len)
{
	const u32 packet_tag = 1;
	u32 total_len, captured_len = data_len;
	int offset;
	int rc;

	/* Header structure, all in host order. */
	struct __attribute__ ((__packed__)) {
		uint32_t total_len;
		uint32_t tag;
		uint32_t original_len;
		uint32_t captured_len;
		uint32_t timestamp;
	} header;

	if (!log_entire_packet(efx->netdev, data))
		captured_len = MIN(captured_len, NVLOG_TRUNCATED_SIZE);

	total_len = sizeof(header) + captured_len;

	offset = efx_nvlog_state.offset;
	if (offset + total_len > efx_nvlog_state.limit) {
		/* Prevent future (shorter) packets being logged. */
		efx_nvlog_state.offset = efx_nvlog_state.limit;
		return;
	}

	/* Prepare header fields:
	 *  total length, including this header
	 *  1 - indicates packet data
	 *  original length of data
	 *  length of data in log - this can be truncated for some packets
	 *  timestamp in microseconds
	 */
	header.total_len = total_len;
	header.tag = packet_tag;
	header.original_len = data_len;
	header.captured_len = captured_len;
	header.timestamp = currticks() * 1000000 / ticks_per_sec();

	efx_nvram_start(efx, efx_nvlog_state.partition);

	rc = efx_nvram_write_locked(efx, efx_nvlog_state.partition, offset,
				    sizeof(header), &header);
	if (rc) {
		DBGC(efx, "NvLog packet header write failed: %s\n",
		     strerror(rc));
		goto out;
	}
	offset += sizeof(header);

	rc = efx_nvram_write_locked(efx, efx_nvlog_state.partition, offset,
				    captured_len, data);
	if (rc) {
		DBGC(efx, "NvLog packet write failed: %s\n", strerror(rc));
		goto out;
	}
	offset += captured_len;

out:
	efx_nvram_finish(efx, efx_nvlog_state.partition);
	efx_nvlog_state.offset = offset;
}

#endif
