/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include "sfc_clp.h"
#include <string.h>
#include <mc_driver_pcol.h>
#include <gpxe/if_ether.h>
#include <gpxe/sf_startup.h>
#include <gpxe/instance.h>

void sfc_init_clp_dev(struct clp_device *clp_dev, struct pci_device *pci)
{
	struct efx_nic *efx = netdev_priv(pci_get_drvdata(pci));

	/* Siena NICs list all ports as netport1.
	 * Huntington NICs list ports increasing from netport1.
	 */
	if (efx->revision == EFX_SIENA)
		clp_dev->netport = 1;
	else if (efx->revision == EFX_HUNTINGTON)
		clp_dev->netport = fn_from_busdevfn(pci->devfn) + 1;
	clp_dev->pci_dev = pci;
	clp_dev->device = efx;
	clp_dev->reset = sfc_clp_reset;
	clp_dev->get_mac = sfc_clp_get_mac;
	clp_dev->set_perm_mac = sfc_clp_set_perm_mac;
	clp_dev->set_boot = sfc_clp_set_boot;
}

void sfc_fini_clp_dev(struct clp_device *clp_dev)
{
	memset(clp_dev, 0, sizeof(*clp_dev));
}

void sfc_clp_reset(struct clp_device *clp_dev, struct clp_response *response)
{
	struct efx_nic *efx = clp_dev->device;
	int rc;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CLP_IN_DEFAULT_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_CLP_OUT_DEFAULT_LEN);

	MCDI_SET_DWORD(inbuf, CLP_IN_OP, MC_CMD_CLP_OP_DEFAULT);
	rc = efx->type->mcdi_rpc(efx, MC_CMD_CLP, inbuf,
				 MC_CMD_CLP_IN_DEFAULT_LEN, outbuf,
				 MC_CMD_CLP_OUT_DEFAULT_LEN, NULL, false);

	if (rc) {
		DBG("Failed CLP reset on CLP device %p (rc 0x%x)\n", clp_dev,
		    rc);
		response->string = CLP_STRING_INTERNAL_ERROR;
		response->status = CLP_STATUS_INTERNAL_ERROR;
	} else {
		response->string = CLP_STRING_SUCCESS;
		response->status = CLP_STATUS_SUCCESS;
	}
}

int sfc_clp_get_mac(struct clp_device *clp_dev, uint8_t *mac)
{
	struct efx_nic *efx = clp_dev->device;
	u8 *mac_ptr;
	int rc;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CLP_IN_GET_MAC_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_CLP_OUT_GET_MAC_LEN);

	MCDI_SET_DWORD(inbuf, CLP_IN_OP, MC_CMD_CLP_OP_GET_MAC);
	rc = efx->type->mcdi_rpc(efx, MC_CMD_CLP, inbuf, sizeof(inbuf),
				 outbuf, sizeof(outbuf), NULL, false);

	if (rc == 0) {
		mac_ptr = MCDI_PTR(outbuf, CLP_OUT_GET_MAC_ADDR);
	} else {
		/* CLP data didn't include the MAC address, so try to get it
		 * from board config.
		 *
		 * Really this should only be done for rc=-EBUSY, but gpxe makes
		 * it more complicated to compare error numbers, so just do this
		 * for now.
		 */
		MCDI_DECLARE_BUF(outbuf2, MC_CMD_GET_BOARD_CFG_OUT_LENMAX);
		rc = efx->type->mcdi_rpc(efx, MC_CMD_GET_BOARD_CFG, NULL, 0,
					 outbuf2, sizeof(outbuf2), NULL,
					 false);

		if (rc)
			goto error;

		mac_ptr = efx->port ?
			MCDI_PTR(outbuf2, GET_BOARD_CFG_OUT_MAC_ADDR_BASE_PORT1) :
			MCDI_PTR(outbuf2, GET_BOARD_CFG_OUT_MAC_ADDR_BASE_PORT0);
	}

	memcpy(mac, mac_ptr, ETH_ALEN);
	return 0;

error:
	DBG("Failed to get MAC address for CLP device %p (rc 0x%x)\n", clp_dev,
	    rc);
	return rc;
}

void sfc_clp_set_perm_mac(struct clp_device *clp_dev,
			  struct clp_response *response, const uint8_t *mac)
{
	struct efx_nic *efx = clp_dev->device;
	int rc;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CLP_IN_SET_MAC_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_CLP_OUT_SET_MAC_LEN);

	memcpy(MCDI_PTR(inbuf, CLP_IN_SET_MAC_ADDR), mac, ETH_ALEN);
	MCDI_SET_DWORD(inbuf, CLP_IN_OP, MC_CMD_CLP_OP_SET_MAC);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_CLP, inbuf, sizeof(inbuf),
				 outbuf, sizeof(outbuf), NULL, false);

	if (rc) {
		DBG("Failed to set permanent MAC address for CLP device %p "
		    "(rc 0x%x)\n", clp_dev, rc);
		response->string = CLP_STRING_INTERNAL_ERROR;
		response->status = CLP_STATUS_INTERNAL_ERROR;
	} else {
		response->string = CLP_STRING_SUCCESS;
		response->status = CLP_STATUS_SUCCESS;
	}
}

void sfc_clp_set_boot(struct clp_device *clp_dev, struct clp_response *response,
		      int enable)
{
	struct efx_nic *efx = clp_dev->device;
	int rc;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CLP_IN_SET_BOOT_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_CLP_OUT_SET_BOOT_LEN);

	MCDI_SET_DWORD(inbuf, CLP_IN_OP, MC_CMD_CLP_OP_SET_BOOT);
	if (enable)
		*MCDI_PTR(inbuf, CLP_IN_SET_BOOT_FLAG) = SF_BOOT_MODE_BEV;
	else
		*MCDI_PTR(inbuf, CLP_IN_SET_BOOT_FLAG) = SF_BOOT_MODE_DISABLE;

	rc = efx->type->mcdi_rpc(efx, MC_CMD_CLP, inbuf, sizeof(inbuf),
				 outbuf, sizeof(outbuf), NULL, false);

	if (rc) {
		DBG("Failed to set boot flag for CLP device %p (rc 0x%x)\n",
		    clp_dev, rc);
		response->string = CLP_STRING_INTERNAL_ERROR;
		response->status = CLP_STATUS_INTERNAL_ERROR;
	} else {
		response->string = CLP_STRING_SUCCESS;
		response->status = CLP_STATUS_SUCCESS;
	}
}
