/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/**************************************************************************
gPXE -  Network Bootstrap Program

Literature dealing with the network protocols:
	ARP - RFC826
	RARP - RFC903
	UDP - RFC768
	BOOTP - RFC951, RFC2132 (vendor extensions)
	DHCP - RFC2131, RFC2132 (options)
	TFTP - RFC1350, RFC2347 (options), RFC2348 (blocksize), RFC2349 (tsize)
	RPC - RFC1831, RFC1832 (XDR), RFC1833 (rpcbind/portmapper)
	NFS - RFC1094, RFC1813 (v3, useful for clarifications, not implemented)
	IGMP - RFC1112

**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gpxe/init.h>
#include <gpxe/banner.h>
#include <gpxe/bcv.h>
#include <gpxe/instance.h>
#include <gpxe/shell.h>
#include <usr/autoboot.h>
#include <usr/bootconnect.h>
#include <usr/prompt.h>
#include <config/general.h>

#include <gpxe/settings.h>

#include <gpxe/sf_startup.h>
#include <gpxe/sf_settings.h>
#include <gpxe/device.h>
#include <gpxe/netdevice.h>
#include <gpxe/if_ether.h>
#include <arch/i386/include/registers.h>

#include <console.h>
#include <unistd.h>
#include <gpxe/keys.h>

/**
 * Prompt for shell entry
 *
 * @ret	enter_shell		User wants to enter shell
 */
static int shell_banner ( void ) {
	struct net_device *netdev;
	struct settings *settings;
	int banner_secs = 5;

	/* Try to get banner delay setting; use the 5s default if it fails */
	netdev = instance_netdevice();
	if ( netdev ) {
		settings = netdev_nvo_settings ( netdev );
		banner_secs = sf_sett_get_intz ( settings,
						 &banner_delay_setting );
		if ( banner_secs <= 0 )
			banner_secs = 5;
	}

	/* Skip prompt if timeout is zero */
	if ( banner_secs == 0 )
		return 0;

	return ( prompt ( "\nPress Ctrl-B for the Solarflare Boot"
			  " Configuration Utility...",
			  ( banner_secs * TICKS_PER_SEC ), CTRL_B ) == 0 );
}

/**
 * Determine boot mode (network boot, iSCSI connect, or disable)
 *
 * @ret boot_mode		Magic value for ROM prefix
 */
static uint32_t select_boot_mode ( struct net_device *netdev ) {
#ifdef BCV_PROTO_ISCSI
	uint64_t		priority, key;
	struct net_device	*nd;
#endif
	struct settings  	*settings;
	unsigned long  		boot_type;

	settings = netdev_nvo_settings ( netdev );
	if (!settings)
		return SF_BOOT_MODE_DISABLE;

	boot_type = sf_sett_get_intz ( settings, &boot_type_setting );

#ifdef BCV_PROTO_ISCSI
	if ( boot_type == BOOTTYPE_ISCSI ) {
		/* Only the highest priority (lowest rank)
		 * mpio device is set to BCV booting */
		priority = sf_mpio_priority_key ( netdev );
		list_for_each_entry ( nd, &net_devices, list ) {
			if ( nd == netdev )
				continue;

			key = sf_mpio_priority_key ( nd );
			if ( key == PRIORITY_KEY_NONE )
				continue;
			if ( key < priority )
				return SF_BOOT_MODE_DISABLE;
		}

		/* Must be the lowest rank */
		return SF_BOOT_MODE_BCV;
	} else if ( boot_type == BOOTTYPE_PXE )
#else
	if ( boot_type == BOOTTYPE_PXE )
#endif

		return SF_BOOT_MODE_BEV;
	else
		return SF_BOOT_MODE_DISABLE;
}

/**
 * Main entry point for init-time mode selection
 *
 * @mode_out	      SF_BOOT_MODE_BCV to switch to iSCSI boot (BCV) mode
 *		      SF_BOOT_MODE_DISABLE to disable booting altogether
 *		      or any other value to remain in PXE boot (BEV) mode
 * @ll_addr_out	      The link local address of the netdevice
 * @bootskip_delay    The prefix will ask for confirmation for @bootskip_delay
 *                    seconds
 */
void init_select_mode ( uint32_t *mode_out, uint8_t *ll_addr_out,
			uint8_t *bootskip_delay ) {
	struct net_device *netdev;

	initialise();
	startup();

	welcome();

	/* Give an opportunity to enter the shell before continuing */
	if ( shell_banner() )
#if defined(SHELL_HIDE) || defined(SHELL_DISABLE)
		sfconfig();
#else
		shell();
#endif

	/* Select boot mode: PXE (BEV), iSCSI (BCV), or disabled */
	netdev = instance_netdevice();
	if ( netdev == NULL ) {
		printf ( "Failed to find boot device\n" );

		*mode_out = SF_BOOT_MODE_DISABLE;
		*bootskip_delay = 0;
		memset ( ll_addr_out, 0, ETH_ALEN );
	} else {
		*mode_out = select_boot_mode ( netdev );

		*bootskip_delay = sf_sett_get_intz ( netdev_nvo_settings ( netdev ),
						     &bootskip_delay_setting );
		memcpy ( ll_addr_out, netdev->ll_addr, ETH_ALEN );
	}

	shutdown_exit();
}

/**
 * Main entry point for BCV
 */
void bcv_main ( void ) {
#ifdef BCV_PROTO_ISCSI
	initialise();
	startup();

	welcome();

#if defined(SHELL_HIDE) || defined(SHELL_DISABLE)
	sf_bootconnect();

	/* Pause briefly so that we can see any error messages */
	mdelay(5000);
#else
	if ( shell_banner() ) {
		/* User wants shell; just give them a shell */
		shell();
	} else {
		/* User doesn't want shell; try booting.  If booting
		 * fails, offer a second chance to enter the shell for
		 * diagnostics.
		 */
		/* OKTL code: use sf_autoboot() */
		/* autoboot(); */
		sf_bootconnect();

		if ( shell_banner() )
			shell();
	}
#endif

	shutdown_exit();
#endif
}
