/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Copyright (C) 2006 Michael Brown <mbrown@fensystems.co.uk>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <byteswap.h>
#include <gpxe/vsprintf.h>
#include <gpxe/socket.h>
#include <gpxe/xfer.h>
#include <gpxe/open.h>
#include <gpxe/scsi.h>
#include <gpxe/process.h>
#include <gpxe/uaccess.h>
#include <gpxe/tcpip.h>
#include <gpxe/settings.h>
#include <gpxe/features.h>
#include <gpxe/iscsi.h>

/** @file
 *
 * iSCSI protocol support
 *
 */

/** iSCSI root path components (as per RFC4173) */
enum iscsi_root_path_component {
       RP_LITERAL = 0,
       RP_SERVERNAME,
       RP_PROTOCOL,
       RP_PORT,
       RP_LUN,
       RP_TARGETNAME,
       NUM_RP_COMPONENTS
};

/**
 * Parse iSCSI root path
 *
 * @v root_path   	iSCSI root path string (as per RFC4173)
 * @v p_servername	Server name string location
 * @v p_port      	Port string location
 * @v p_lun       	LUN string location
 * @v p_iqn       	Target IQN string location
 *
 * @ret rc        	Return status code
 */
int iscsi_parse_root_path ( const char *root_path,
			    char **p_servername, unsigned int *p_port,
			    uint64_t *p_lun, char **p_iqn ) {
	char *rp_copy;
	char *rp_comp[NUM_RP_COMPONENTS];
	char *rp;
	int i = 0;
	char *serv = NULL;
	char *iqn = NULL;
	int rc = 0;

	/* Split root path into component parts */
	if ( ( rp_copy = strdup ( root_path ) ) == NULL )
		goto err_root;
	rp = rp_copy;

	while ( 1 ) {
		rp_comp[i++] = rp;
		if ( i == NUM_RP_COMPONENTS )
			break;
		for ( ; *rp != ':' ; rp++ ) {
			if ( ! *rp ) {
				DBG ( "iSCSI root path \"%s\" "
				      "too short\n", root_path );
				rc = -EINVAL;
				goto err_short;
			}
		}
		*(rp++) = '\0';
	}

	if ( p_servername != NULL ) {
		if ( (serv = strdup ( rp_comp[RP_SERVERNAME] )) == NULL ) {
			rc = -ENOMEM;
			goto err_serv;
		}
		*p_servername = serv;
	}
	if ( p_port != NULL ) {
		*p_port = ISCSI_PORT;
		if ( rp_comp[RP_PORT][0] != '\0' )
			*p_port = strtoul ( rp_comp[RP_PORT], NULL, 10 );
	}

	if ( p_lun != NULL )
		/* We don't support LUN > 256 in the DHCP response,
		 * so no need to use strtoull here. LUN is hex */
		*p_lun = strtoul ( rp_comp[RP_LUN], NULL, 16 );

	if ( p_iqn != NULL ) {
		if ( (iqn = strdup ( rp_comp[RP_TARGETNAME] )) == NULL ) {
			rc = -ENOMEM;
			goto err_iqn;
		}
		*p_iqn = iqn;
	}

	free ( rp_copy );
	return 0;

err_iqn:
	free ( serv );
err_serv:
err_short:
	free ( rp_copy );
err_root:
	return rc;
}
